#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['GracefulCoroDestroy', 'graceful_coro_destroyer', 'agraceful_coro_destroyer', 'set_my_exit_handler', 'set_exit_handler', 'kill_coro_children', 'kill_my_children_on_exit', 'kill_children_on_exit']


from typing import Optional, Type, Any, Hashable, cast, Set, Callable
from cengal.parallel_execution.coroutines.coro_scheduler import CoroScheduler, CoroSchedulerType, Interface, CoroID, ExplicitWorker, Worker, CoroWrapperBase, get_interface_for_an_explicit_loop, current_interface, current_coro_scheduler
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import PutCoro, put_coro_to, put_coro, get_set_of_all_coro_children
from cengal.parallel_execution.coroutines.coro_standard_services.sleep import Sleep
from cengal.parallel_execution.coroutines.coro_standard_services.simple_yield import Yield
from cengal.parallel_execution.coroutines.coro_standard_services.kill_coro import KillCoro, kill_coro
from cengal.parallel_execution.coroutines.coro_standard_services.kill_coro_list import KillCoroList, kill_coro_list
from cengal.parallel_execution.coroutines.coro_standard_services.throw_coro import ThrowCoro
from cengal.parallel_execution.coroutines.coro_standard_services.wait_coro import WaitCoro, WaitCoroRequest, CoroutineNotFoundError
from cengal.code_flow_control.smart_values import ValueExistence


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class GracefulCoroDestroy(Exception):
    pass


def graceful_coro_destroyer(
        i: Interface, 
        phase_time_limit: Optional[float], 
        coro_id: CoroID, 
        ex_type: Type[Exception] = None, ex_value: Exception = None, ex_traceback: Any = None, 
        tree: bool = True, 
        first_phase_is_wait: bool = True, 
        last_phase_is_kill: bool = True, 
    ):
    phase_time_limit = phase_time_limit or 0
    try:
        if first_phase_is_wait:
            try:
                i(WaitCoro, WaitCoroRequest(phase_time_limit, kill_on_timeout=False, tree=tree, result_required=False).single(coro_id))
            except TimeoutError:
                pass

        if (ex_type is not None) or (ex_value is not None):
            if not i(ThrowCoro, coro_id, ex_type, ex_value, ex_traceback, tree):
                raise CoroutineNotFoundError
            
            try:
                i(WaitCoro, WaitCoroRequest(phase_time_limit, kill_on_timeout=False, tree=tree, result_required=False).single(coro_id))
            except TimeoutError:
                pass

        if not i(ThrowCoro, coro_id, GracefulCoroDestroy, tree=tree):
            raise CoroutineNotFoundError
        
        try:
            i(WaitCoro, WaitCoroRequest(phase_time_limit, kill_on_timeout=last_phase_is_kill, tree=tree, result_required=False).single(coro_id))
        except TimeoutError:
            pass
        
    except CoroutineNotFoundError:
        pass
    finally:
        i(Yield)


async def agraceful_coro_destroyer(
        i: Interface, 
        phase_time_limit: Optional[float], 
        coro_id: CoroID, 
        ex_type: Type[Exception] = None, ex_value: Exception = None, ex_traceback: Any = None, 
        tree: bool = True, 
        first_phase_is_wait: bool = True,
        last_phase_is_kill: bool = True, 
    ):
    phase_time_limit = phase_time_limit or 0
    try:
        if first_phase_is_wait:
            try:
                await i(WaitCoro, WaitCoroRequest(phase_time_limit, kill_on_timeout=False, tree=tree, result_required=False).single(coro_id))
            except TimeoutError:
                pass

        if (ex_type is not None) or (ex_value is not None):
            if not await i(ThrowCoro, coro_id, ex_type, ex_value, ex_traceback, tree):
                raise CoroutineNotFoundError
            
            try:
                await i(WaitCoro, WaitCoroRequest(phase_time_limit, kill_on_timeout=False, tree=tree, result_required=False).single(coro_id))
            except TimeoutError:
                pass

        if not await i(ThrowCoro, coro_id, GracefulCoroDestroy, tree=tree):
            raise CoroutineNotFoundError
        
        try:
            await i(WaitCoro, WaitCoroRequest(phase_time_limit, kill_on_timeout=last_phase_is_kill, tree=tree, result_required=False).single(coro_id))
        except TimeoutError:
            pass
        
    except CoroutineNotFoundError:
        pass
    finally:
        await i(Yield)


def set_my_exit_handler(handler: Callable[[CoroWrapperBase], Optional[bool]]) -> bool:
    i: Interface = current_interface()
    coro = cast(CoroWrapperBase, i._coro)
    coro.add_on_coro_del_handler(handler)


def set_exit_handler(handler: Callable[[CoroWrapperBase], Optional[bool]], coro_id: Optional[CoroID] = None) -> bool:
    current_loop: CoroSchedulerType = current_coro_scheduler()
    if coro_id is None:
        i: Interface = current_loop.current_interface()
        if i is None:
            return False
        
        coro_id = i.coro_id
    
    if coro_id is None:
        return False
    
    coro, was_new_born, new_born_index = current_loop.find_coro_by_id(coro_id)
    coro = cast(CoroWrapperBase, coro)
    coro.add_on_coro_del_handler(handler)


def kill_coro_children(coro: CoroWrapperBase):
    coro_id = coro.coro_id
    children: Set[CoroID] = get_set_of_all_coro_children(coro_id)
    for child_id in children:
        kill_coro(child_id, tree=True)


def kill_my_children_on_exit() -> bool:
    i: Interface = current_interface()
    coro = cast(CoroWrapperBase, i._coro)
    coro.add_on_coro_del_handler(kill_coro_children)


def kill_children_on_exit(coro_id: Optional[CoroID] = None) -> bool:
    current_loop: CoroSchedulerType = current_coro_scheduler()
    if coro_id is None:
        i: Interface = current_loop.current_interface()
        if i is None:
            return False
        
        coro_id = i.coro_id
    
    if coro_id is None:
        return False
    
    coro, was_new_born, new_born_index = current_loop.find_coro_by_id(coro_id)
    coro = cast(CoroWrapperBase, coro)
    coro.add_on_coro_del_handler(kill_coro_children)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['is_mobile_browser', 'is_tablet_browser', 'is_mobile_or_tablet_browser']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


import re


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


reg_b = re.compile('(android|bb\\\\d+|meego).+mobile|avantgo|bada\\\\/|blackberry|blazer|compal|elaine|fennec|hiptop|iemobile|ip(hone|od)|iris|kindle|lge |maemo|midp|mmp|mobile.+firefox|netfront|opera m(ob|in)i|palm( os)?|phone|p(ixi|re)\\\\/|plucker|pocket|psp|series(4|6)0|symbian|treo|up\\\\.(browser|link)|vodafone|wap|windows ce|xda|xiino', re.I | re.M)
reg_v = re.compile('1207|6310|6590|3gso|4thp|50[1-6]i|770s|802s|a wa|abac|ac(er|oo|s\\\\-)|ai(ko|rn)|al(av|ca|co)|amoi|an(ex|ny|yw)|aptu|ar(ch|go)|as(te|us)|attw|au(di|\\\\-m|r |s )|avan|be(ck|ll|nq)|bi(lb|rd)|bl(ac|az)|br(e|v)w|bumb|bw\\\\-(n|u)|c55\\\\/|capi|ccwa|cdm\\\\-|cell|chtm|cldc|cmd\\\\-|co(mp|nd)|craw|da(it|ll|ng)|dbte|dc\\\\-s|devi|dica|dmob|do(c|p)o|ds(12|\\\\-d)|el(49|ai)|em(l2|ul)|er(ic|k0)|esl8|ez([4-7]0|os|wa|ze)|fetc|fly(\\\\-|_)|g1 u|g560|gene|gf\\\\-5|g\\\\-mo|go(\\\\.w|od)|gr(ad|un)|haie|hcit|hd\\\\-(m|p|t)|hei\\\\-|hi(pt|ta)|hp( i|ip)|hs\\\\-c|ht(c(\\\\-| |_|a|g|p|s|t)|tp)|hu(aw|tc)|i\\\\-(20|go|ma)|i230|iac( |\\\\-|\\\\/)|ibro|idea|ig01|ikom|im1k|inno|ipaq|iris|ja(t|v)a|jbro|jemu|jigs|kddi|keji|kgt( |\\\\/)|klon|kpt |kwc\\\\-|kyo(c|k)|le(no|xi)|lg( g|\\\\/(k|l|u)|50|54|\\\\-[a-w])|libw|lynx|m1\\\\-w|m3ga|m50\\\\/|ma(te|ui|xo)|mc(01|21|ca)|m\\\\-cr|me(rc|ri)|mi(o8|oa|ts)|mmef|mo(01|02|bi|de|do|t(\\\\-| |o|v)|zz)|mt(50|p1|v )|mwbp|mywa|n10[0-2]|n20[2-3]|n30(0|2)|n50(0|2|5)|n7(0(0|1)|10)|ne((c|m)\\\\-|on|tf|wf|wg|wt)|nok(6|i)|nzph|o2im|op(ti|wv)|oran|owg1|p800|pan(a|d|t)|pdxg|pg(13|\\\\-([1-8]|c))|phil|pire|pl(ay|uc)|pn\\\\-2|po(ck|rt|se)|prox|psio|pt\\\\-g|qa\\\\-a|qc(07|12|21|32|60|\\\\-[2-7]|i\\\\-)|qtek|r380|r600|raks|rim9|ro(ve|zo)|s55\\\\/|sa(ge|ma|mm|ms|ny|va)|sc(01|h\\\\-|oo|p\\\\-)|sdk\\\\/|se(c(\\\\-|0|1)|47|mc|nd|ri)|sgh\\\\-|shar|sie(\\\\-|m)|sk\\\\-0|sl(45|id)|sm(al|ar|b3|it|t5)|so(ft|ny)|sp(01|h\\\\-|v\\\\-|v )|sy(01|mb)|t2(18|50)|t6(00|10|18)|ta(gt|lk)|tcl\\\\-|tdg\\\\-|tel(i|m)|tim\\\\-|t\\\\-mo|to(pl|sh)|ts(70|m\\\\-|m3|m5)|tx\\\\-9|up(\\\\.b|g1|si)|utst|v400|v750|veri|vi(rg|te)|vk(40|5[0-3]|\\\\-v)|vm40|voda|vulc|vx(52|53|60|61|70|80|81|83|85|98)|w3c(\\\\-| )|webc|whit|wi(g |nc|nw)|wmlb|wonu|x700|yas\\\\-|your|zeto|zte\\\\-', re.I | re.M)


def is_mobile_browser(user_agent: str) -> bool:
    b = reg_b.search(user_agent)
    v = reg_v.search(user_agent[0:4])
    # print(f"is_mobile_browser. b: {b}; v: {v}.")
    return bool(b or v)


def is_tablet_browser(user_agent: str) -> bool:
    if not is_mobile_browser(user_agent):
        additional_words = ('android', 'touch', 'ipad')
        uac = user_agent.casefold()
        for additional_word in additional_words:
            if additional_word in uac:
                return True

    return False


def is_mobile_or_tablet_browser(user_agent: str) -> bool:
    mobile = is_mobile_browser(user_agent)
    if not mobile:
        additional_words = ('android', 'touch', 'ipad')
        uac = user_agent.casefold()
        for additional_word in additional_words:
            if additional_word in uac:
                return False, True

    return mobile, False



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['copy__impl', 'CopyMixin', 'deepcopy__impl', 'DeepCopyMixin', 'CopyMethodsMixin', 'CopyableMixin', 'get_dict_key_with_callable_default']

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from copy import deepcopy
from typing import Callable, Dict, Hashable, Any
from cengal.data_manipulation.get_dict_key_with_callable_default import get_dict_key_with_callable_default


def copy__impl(self):
    cls = self.__class__
    result = cls.__new__(cls)
    try:
        result.__dict__.update(self.__dict__)
        return result
    except AttributeError:
        pass

    for field in self.__slots__:
        setattr(result, field, getattr(self, field))
    
    return result


class CopyMixin:
    def __copy__(self):
        return copy__impl(self)


def deepcopy__impl(self, memo):
    cls = self.__class__
    result = cls.__new__(cls)
    memo[id(self)] = result
    try:
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        
        return result
    except AttributeError:
        pass

    for field in self.__slots__:
        setattr(result, field, getattr(self, field))
    
    return result


class DeepCopyMixin:
    def __deepcopy__(self, memo):
        return deepcopy__impl(self, memo)



class CopyMethodsMixin:
    def copy(self):
        """Should make relevant copy of an object (not so general and deep as a deepcopy()). should copy only known object fields.
        Example:
            def copy(self):
                cls = self.__class__
                result = cls.__new__(cls)
                result.__dict__['dimension'] = self.dimension
                result.__dict__['_point'] = self._point.copy()
                return result

        Raises:
            NotImplementedError: _description_
        """
        raise NotImplementedError

    def shallow_copy(self):
        """Same as copy.copy(self), but should copy only known object fields.
        Example:
            def shallow_copy(self):
                cls = self.__class__
                result = cls.__new__(cls)
                result.__dict__['dimension'] = self.dimension
                result.__dict__['_point'] = self._point
                return result

        Raises:
            NotImplementedError: _description_
        """
        raise NotImplementedError

    def updated_copy(self, update: Dict):
        """Will make updated copy of an object. Other behavior should be the same as in the `def copy(self)` method.
        Example:
            # from cengal.data_manipulation.get_dict_key_with_callable_default import get_dict_key_with_callable_default
            from cengal.entities.copyable import CopyableMixin, get_dict_key_with_callable_default
            
            def updated_copy(self, update: Dict):
                cls = self.__class__
                result = cls.__new__(cls)
                result.__dict__['dimension'] = update.get('dimension', self.dimension)
                result.__dict__['_point'] = get_dict_key_with_callable_default(update, '_point', lambda: self._point.copy())
                return result

        Raises:
            NotImplementedError: _description_
        """
        raise NotImplementedError


class CopyableMixin(CopyMixin, DeepCopyMixin, CopyMethodsMixin):
    pass



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    'append_cflags', 
    'concat_cflags', 
    'wrap_definition_text', 
    'wrap_definition_text_raw', 
    'prepare_definition_value', 
    'prepare_definition_value_raw', 
    'convert_value_to_flag', 
    'convert_value_to_text', 
    'prepare_definition', 
    'list_to_dict', 
    'dict_of_tuples_to_dict', 
    'adjust_definition_names', 
    'prepare_cflags', 
    'prepare_given_cflags', 
    'prepare_cflags_dict', 
    'prepare_given_cflags_dict', 
    'prepare_pyx_flags_dict', 
    'prepare_compile_time_env', 
    'prepare_compile_time_flags',
    ]


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from typing import Dict, Any, Tuple, Union
from os import environ
try:
    from os import uname
    uname_sysname: str = uname().sysname
    uname_machine: str = uname().machine
except ImportError:
    from platform import uname
    uname_sysname: str = uname().system
    uname_machine: str = uname().machine

from cengal.hardware.info.cpu import cpu_info
from cengal.os.execute import escape_text, escape_param
from cengal.build_tools.current_compiler import compiler_type, compiler_name, compiler_string
from cengal.system import (
    PYTHON_IMPLEMENTATION,
    PYTHON_VERSION,
    PYTHON_VERSION_STR,
    PYTHON_VERSION_INT,
    IS_RUNNING_IN_PYCHARM,
    RAW_OS_PLATFORM,
    OS_API_TYPE,
    OS_TYPE,
    KIVY_PLATFORM,
    KIVY_TARGET_PLATFORM,
    IS_RUNNING_IN_EMSCRIPTEN,
    IS_RUNNING_IN_WASI,
    IS_RUNNING_IN_PYODIDE,
    IS_BUILDING_FOR_PYODIDE,
    IS_INSIDE_OR_FOR_WEB_BROWSER,
    CENGAL_VERSION_STR,
    CENGAL_VERSION_MAJOR_STR,
    CENGAL_VERSION_MINOR_STR,
    CENGAL_VERSION_MICRO_STR,
    CENGAL_VERSION_MAJOR,
    CENGAL_VERSION_MINOR,
    CENGAL_VERSION_MICRO,
)
from typing import Set, Sequence, Optional, Sequence, List


def append_cflags(cflags_list: list):
    if 'msvc' == compiler_type:
        compiler_flags_env_var_name = 'CL'
    else:
        compiler_flags_env_var_name = 'CFLAGS'

    environ[compiler_flags_env_var_name] = ' '.join((environ.get(compiler_flags_env_var_name, str()), " ".join(cflags_list)))


def concat_cflags(cflags: str):
    if 'msvc' == compiler_type:
        compiler_flags_env_var_name = 'CL'
    else:
        compiler_flags_env_var_name = 'CFLAGS'

    environ[compiler_flags_env_var_name] = ' '.join((environ.get(compiler_flags_env_var_name, str()), cflags))


def wrap_definition_text(text: str) -> str:
    if 'msvc' == compiler_type:
        return escape_param(f'\\"{text}\\"')
    else:
        return f'\\"{escape_param(text)}\\"'


def wrap_definition_text_raw(text: str) -> str:
    return f'"{text}"'


def wrap_pyx_definition_text_raw(text: str) -> str:
    return f'{text}'


def prepare_definition_value(value: Union[None, bool, int, str]) -> str:
    if isinstance(value, bool):
        value = str(value)
    
    if value is None:
        return None
    elif isinstance(value, int):
        return value
    elif isinstance(value, str):
        return wrap_definition_text(value)
    else:
        return wrap_definition_text(f'{value}')


def prepare_definition_value_raw(value: Union[None, bool, int, str]) -> str:
    if isinstance(value, bool):
        value = str(value)
    
    if value is None:
        return None
    elif isinstance(value, int):
        return value
    elif isinstance(value, str):
        return wrap_definition_text_raw(value)
    else:
        return wrap_definition_text_raw(f'{value}')


def prepare_pyx_definition_value_raw(value: Union[None, bool, int, str]) -> str:
    if isinstance(value, bool):
        value = str(value)
    
    if value is None:
        return None
    elif isinstance(value, int):
        return value
    elif isinstance(value, str):
        return wrap_pyx_definition_text_raw(value)
    else:
        return wrap_pyx_definition_text_raw(f'{value}')


def wrap_definition_pair(name: str, value: Any) -> str:
    if 'msvc' == compiler_type:
        return f'/D{name}={value}' if value is not None else f'/D{name}'
    else:
        return f'-D{name}={value}' if value is not None else f'-D{name}'


def convert_value_to_flag(value: Union[None, bool, int, str]) -> int:
    if value is None:
        return None
    
    try:
        return int(value)
    except:
        return 1


def convert_value_to_text(value: Union[None, bool, int, str]) -> str:
    return f'{value}'


def prepare_definition(name: str, value: Optional[Union[bool, int, str]] = None) -> str:
    return wrap_definition_pair(name, prepare_definition_value(value))


def list_to_dict(additional_cflags: Optional[Sequence[str]] = None) -> Dict[str, Union[None, bool, int, str]]:
    additional_cflags = dict() if additional_cflags is None else additional_cflags
    if not isinstance(additional_cflags, dict):
        additional_cflags = {name: None for name in additional_cflags}
    
    return additional_cflags


def dict_of_tuples_to_dict(additional_cflags: Optional[Dict[str, Tuple[bool, Union[None, bool, str, int]]]] = None) -> Dict[str, Union[None, bool, int, str]]:
    additional_cflags = dict() if additional_cflags is None else additional_cflags
    adjusted_additional_cflags: Dict[str, Union[None, bool, int, str]] = dict()
    for name, value_info in additional_cflags.items():
        if isinstance(value_info, tuple):
            is_flag, value = value_info
            value = convert_value_to_flag(value) if is_flag else convert_value_to_text(value)
        else:
            value = value_info
        
        adjusted_additional_cflags[name] = value
    
    return adjusted_additional_cflags


def adjust_definition_names(definitions: Dict[str, Union[None, bool, int, str]], flag_name_prefix: str = None, definition_name_prefix: str = None) -> Dict[str, Union[None, bool, int, str]]:
    flag_name_prefix = str() if flag_name_prefix is None else flag_name_prefix
    definition_name_prefix = str() if definition_name_prefix is None else definition_name_prefix
    adjusted_definitions: Dict[str, Union[None, bool, int, str]] = dict()
    for name, value in definitions.items():
        if value is None:
            name = f'{flag_name_prefix}{name}'
        else:
            name = f'{definition_name_prefix}{name}'
        
        adjusted_definitions[name] = value
    
    return adjusted_definitions


def prepare_cflags(additional_cflags: Optional[Union[Sequence[str], Dict[str, Union[Union[None, bool, int, str], Tuple[bool, Union[None, bool, str, int]]]]]] = None) -> List[Union[None, int, str]]:
    adjusted_additional_cflags: Dict[str, Union[None, bool, int, str]] = dict_of_tuples_to_dict(list_to_dict(additional_cflags))
    flags: Dict[str, None] = adjust_definition_names(list_to_dict(prepare_compile_time_flags()), 'CF_', 'CD_')  # CF is Cengal Flag; CD is Cengal Definition
    definitions: Dict[str, Union[None, bool, int, str]] = adjust_definition_names(prepare_compile_time_env(), 'CF_', 'CD_')  # CF is Cengal Flag; CD is Cengal Definition
    definitions.update(flags)
    definitions.update(adjusted_additional_cflags)
    params: List[Union[None, int, str]] = [prepare_definition(name, value) for name, value in definitions.items()]
    
    # from pprint import pprint
    # print('<<< PREPARE_CFLAGS: >>>')
    # pprint(params)
    append_cflags(params)
    return params


def prepare_given_cflags(additional_cflags: Optional[Union[Sequence[str], Dict[str, Union[Union[None, bool, int, str], Tuple[bool, Union[None, bool, str, int]]]]]] = None) -> List[Union[None, int, str]]:
    adjusted_additional_cflags: Dict[str, Union[None, bool, int, str]] = dict_of_tuples_to_dict(list_to_dict(additional_cflags))
    params: List[Union[None, int, str]] = [prepare_definition(name, value) for name, value in adjusted_additional_cflags.items()]
    
    # from pprint import pprint
    # print('<<< PREPARE_CFLAGS: >>>')
    # pprint(params)
    append_cflags(params)
    return params


def prepare_cflags_dict(additional_cflags: Optional[Union[Sequence[str], Dict[str, Union[Union[None, bool, int, str], Tuple[bool, Union[None, bool, str, int]]]]]] = None) -> Dict[str, Union[None, bool, int, str]]:
    adjusted_additional_cflags: Dict[str, Union[None, bool, int, str]] = dict_of_tuples_to_dict(list_to_dict(additional_cflags))
    flags: Dict[str, None] = adjust_definition_names(list_to_dict(prepare_compile_time_flags()), 'CF_', 'CD_')  # CF is Cengal Flag; CD is Cengal Definition
    definitions: Dict[str, Union[None, bool, int, str]] = adjust_definition_names(prepare_compile_time_env(), 'CF_', 'CD_')  # CF is Cengal Flag; CD is Cengal Definition
    definitions.update(flags)
    definitions.update(adjusted_additional_cflags)
    result: Dict[str, Union[None, bool, int, str]] = {name: prepare_definition_value_raw(value) for name, value in definitions.items()}
    
    # from pprint import pprint
    # print('<<< PREPARE_CFLAGS_DICT: >>>')
    # pprint(result)
    return result


def prepare_given_cflags_dict(additional_cflags: Optional[Union[Sequence[str], Dict[str, Union[Union[None, bool, int, str], Tuple[bool, Union[None, bool, str, int]]]]]] = None) -> Dict[str, Union[None, bool, int, str]]:
    adjusted_additional_cflags: Dict[str, Union[None, bool, int, str]] = dict_of_tuples_to_dict(list_to_dict(additional_cflags))
    result: Dict[str, Union[None, bool, int, str]] = {name: prepare_definition_value_raw(value) for name, value in adjusted_additional_cflags.items()}
    
    # from pprint import pprint
    # print('<<< PREPARE_CFLAGS_DICT: >>>')
    # pprint(result)
    return result


def prepare_pyx_flags_dict(additional_pyx_flags: Optional[Union[Sequence[str], Dict[str, Union[Union[None, bool, int, str], Tuple[bool, Union[None, bool, str, int]]]]]] = None) -> Dict[str, Union[None, bool, int, str]]:
    adjusted_additional_cflags: Dict[str, Union[None, bool, int, str]] = dict_of_tuples_to_dict(list_to_dict(additional_pyx_flags))
    flags: Dict[str, None] = adjust_definition_names(list_to_dict(prepare_compile_time_flags()), 'PYXF_', 'PYXD_')  # PYXF is Cython Flag; PYXD is Cython Definition
    definitions: Dict[str, Union[None, bool, int, str]] = adjust_definition_names(prepare_compile_time_env(), 'PYXF_', 'PYXD_')  # PYXF is Cython Flag; PYXD is Cython Definition
    definitions.update(flags)
    definitions.update(adjusted_additional_cflags)
    result: Dict[str, Union[None, bool, int, str]] = {name: prepare_pyx_definition_value_raw(value) for name, value in definitions.items()}
    
    # from pprint import pprint
    # print('<<< PREPARE_CFLAGS_DICT: >>>')
    # pprint(result)
    return result


def prepare_compile_time_env(additional_cflags: Optional[Union[Sequence[str], Dict[str, Union[Union[None, bool, int, str], Tuple[bool, Union[None, bool, str, int]]]]]] = None):
    result = {
        'UNAME_SYSNAME': uname_sysname,  # see: https://en.wikipedia.org/wiki/Uname
        'UNAME_MACHINE': uname_machine,  # see: https://en.wikipedia.org/wiki/Uname
        'CPU_ARCH': cpu_info().arch,  # 'X86_32', 'X86_64', 'ARM_8', 'ARM_7', 'PPC_32', 'PPC_64', 'SPARC_32', 'SPARC_64', 'S390X', 'MIPS_32', 'MIPS_64', 'RISCV_32', 'RISCV_64'
        'CPU_ARCH_RAW_STRING': cpu_info().arch_string_raw,
        'CPU_BITS': cpu_info().bits,  # '32', '64'
        'IS_X86_BOOL': cpu_info().is_x86,  # 'True', 'False'
        'IS_ARM_BOOL': cpu_info().is_arm,  # 'True', 'False'
        'IS_X86': int(cpu_info().is_x86),  # '1', '0'
        'IS_ARM': int(cpu_info().is_arm),  # '1', '0'
        'COMPILER_TYPE': compiler_type,  # 'gcc', 'msvc', 'clang', 'icc', 'llvm', 'intel', 'arm', 'mingw', 'unknown'
        'COMPILER_NAME': compiler_name,  # 'x86_64-linux-gnu-gcc'
        'COMPILER_STRING': compiler_string,  # 'x86_64-linux-gnu-gcc -pthread'
        'PYTHON_IMPLEMENTATION': PYTHON_IMPLEMENTATION,  # 'CPython', 'PyPy', 'IronPython', 'Jython'
        'PYTHON_VERSION_STR': PYTHON_VERSION_STR,  # '3.5.1', ...
        'PYTHON_VERSION_MAJOR_STR': str(PYTHON_VERSION_INT.major),  # '3', ...
        'PYTHON_VERSION_MINOR_STR': str(PYTHON_VERSION_INT.minor),  # '5', ...
        'PYTHON_VERSION_MICRO_STR': str(PYTHON_VERSION_INT.micro),  # '1', ...
        'PYTHON_VERSION_MAJOR': PYTHON_VERSION_INT.major,  # '3', ...
        'PYTHON_VERSION_MINOR': PYTHON_VERSION_INT.minor,  # '5', ...
        'PYTHON_VERSION_MICRO': PYTHON_VERSION_INT.micro,  # '1', ...
        'IS_RUNNING_IN_PYCHARM_BOOL': IS_RUNNING_IN_PYCHARM,  # 'True', 'False'
        'IS_RUNNING_IN_PYCHARM': int(IS_RUNNING_IN_PYCHARM),  # '1', '0'
        'RAW_OS_PLATFORM': RAW_OS_PLATFORM,  # 'emscripten', 'wasi', 'darwin', 'win32', 'cygwin', 'linux', 'linux2', 'linux3', 'darwin', 'freebsd8', 'aix', aix5', 'aix7', ...
        'OS_API_TYPE': OS_API_TYPE,  # 'posix', 'nt', 'java'. Android and iOS will return 'posix'.
        'OS_TYPE': OS_TYPE,  # 'Linux', 'Windows', 'Darwin'
        'KIVY_PLATFORM': KIVY_PLATFORM, # 'win', 'linux', 'android', 'macosx', 'ios', 'unknown'
        'KIVY_TARGET_PLATFORM': KIVY_TARGET_PLATFORM, # 'win', 'linux', 'android', 'macosx', 'ios', 'unknown'
        'IS_RUNNING_IN_EMSCRIPTEN_BOOL': IS_RUNNING_IN_EMSCRIPTEN,  # 'True', 'False'
        'IS_RUNNING_IN_WASI_BOOL': IS_RUNNING_IN_WASI,  # 'True', 'False'
        'IS_RUNNING_IN_PYODIDE_BOOL': IS_RUNNING_IN_PYODIDE,  # 'True', 'False'
        'IS_BUILDING_FOR_PYODIDE_BOOL': IS_BUILDING_FOR_PYODIDE,  # 'True', 'False'
        'IS_INSIDE_OR_FOR_WEB_BROWSER_BOOL': IS_INSIDE_OR_FOR_WEB_BROWSER,  # 'True', 'False'
        'IS_RUNNING_IN_EMSCRIPTEN': int(IS_RUNNING_IN_EMSCRIPTEN),  # '1', '0'
        'IS_RUNNING_IN_WASI': int(IS_RUNNING_IN_WASI),  # '1', '0'
        'IS_RUNNING_IN_PYODIDE': int(IS_RUNNING_IN_PYODIDE),  # '1', '0'
        'IS_BUILDING_FOR_PYODIDE': int(IS_BUILDING_FOR_PYODIDE),  # '1', '0'
        'IS_INSIDE_OR_FOR_WEB_BROWSER': int(IS_INSIDE_OR_FOR_WEB_BROWSER),  # '1', '0'
        'CENGAL_VERSION_STR': CENGAL_VERSION_STR,
        'CENGAL_VERSION_MAJOR_STR': CENGAL_VERSION_MAJOR_STR,
        'CENGAL_VERSION_MINOR_STR': CENGAL_VERSION_MINOR_STR,
        'CENGAL_VERSION_MICRO_STR': CENGAL_VERSION_MICRO_STR,
        'CENGAL_VERSION_MAJOR': CENGAL_VERSION_MAJOR,
        'CENGAL_VERSION_MINOR': CENGAL_VERSION_MINOR,
        'CENGAL_VERSION_MICRO': CENGAL_VERSION_MICRO,
    }
    result.update(dict_of_tuples_to_dict(list_to_dict(additional_cflags)))
    return result


def prepare_compile_time_flags(additional_cflags: Sequence[str] = None):
    additional_cflags = [] if additional_cflags is None else additional_cflags
    macro_flags_list: Set = set()

    if 32 == cpu_info().bits:
        macro_flags_list.add('32_bit_CPU')
    
    if 64 == cpu_info().bits:
        macro_flags_list.add('64_bit_CPU')

    if cpu_info().is_x86:
        macro_flags_list.add('IS_X86')
    
    if cpu_info().is_arm:
        macro_flags_list.add('IS_ARM')

    if 'X86_32' == cpu_info().arch:
        macro_flags_list.add('X86_32')
    elif 'X86_64' == cpu_info().arch:
        macro_flags_list.add('X86_64')
    elif 'ARM_8' == cpu_info().arch:
        macro_flags_list.add('ARM_8')
    elif 'ARM_7' == cpu_info().arch:
        macro_flags_list.add('ARM_7')
    elif 'PPC_32' == cpu_info().arch:
        macro_flags_list.add('PPC_32')
    elif 'PPC_64' == cpu_info().arch:
        macro_flags_list.add('PPC_64')
    elif 'SPARC_32' == cpu_info().arch:
        macro_flags_list.add('SPARC_32')
    elif 'SPARC_64' == cpu_info().arch:
        macro_flags_list.add('SPARC_64')
    elif 'S390X' == cpu_info().arch:
        macro_flags_list.add('S390X')
    elif 'MIPS_32' == cpu_info().arch:
        macro_flags_list.add('MIPS_32')
    elif 'MIPS_64' == cpu_info().arch:
        macro_flags_list.add('MIPS_64')
    elif 'RISCV_32' == cpu_info().arch:
        macro_flags_list.add('RISCV_32')
    elif 'RISCV_64' == cpu_info().arch:
        macro_flags_list.add('RISCV_64')
    
    if 'CPython' == PYTHON_IMPLEMENTATION:
        macro_flags_list.add('CPython')
    elif 'PyPy' == PYTHON_IMPLEMENTATION:
        macro_flags_list.add('PyPy')
    elif 'IronPython' == PYTHON_IMPLEMENTATION:
        macro_flags_list.add('IronPython')
    elif 'Jython' == PYTHON_IMPLEMENTATION:
        macro_flags_list.add('Jython')
    
    if 'gcc' == compiler_type:
        macro_flags_list.add('gcc')
    elif 'msvc' == compiler_type:
        macro_flags_list.add('msvc')
    elif 'clang' == compiler_type:
        macro_flags_list.add('clang')
    elif 'icc' == compiler_type:
        macro_flags_list.add('icc')
    elif 'llvm' == compiler_type:
        macro_flags_list.add('llvm')
    elif 'intel' == compiler_type:
        macro_flags_list.add('intel')
    elif 'arm' == compiler_type:
        macro_flags_list.add('arm')
    elif 'mingw' == compiler_type:
        macro_flags_list.add('mingw')
    elif 'unknown' == compiler_type:
        macro_flags_list.add('unknown_compiler')
    
    if 'Linux' == OS_TYPE:
        macro_flags_list.add('Linux')
    elif 'Windows' == OS_TYPE:
        macro_flags_list.add('Windows')
    elif 'Darwin' == OS_TYPE:
        macro_flags_list.add('Darwin')
    elif 'Java' == OS_TYPE:
        macro_flags_list.add('Java')
    
    if 'win' == KIVY_PLATFORM:
        macro_flags_list.add('KIVY_PLATFORM_win')
    elif 'linux' == KIVY_PLATFORM:
        macro_flags_list.add('KIVY_PLATFORM_linux')
    elif 'android' == KIVY_PLATFORM:
        macro_flags_list.add('KIVY_PLATFORM_android')
    elif 'macosx' == KIVY_PLATFORM:
        macro_flags_list.add('KIVY_PLATFORM_macosx')
    elif 'ios' == KIVY_PLATFORM:
        macro_flags_list.add('KIVY_PLATFORM_ios')
    elif 'unknown' == KIVY_PLATFORM:
        macro_flags_list.add('KIVY_PLATFORM_unknown')
    
    if 'win' == KIVY_TARGET_PLATFORM:
        macro_flags_list.add('KIVY_TARGET_PLATFORM_win')
    elif 'linux' == KIVY_TARGET_PLATFORM:
        macro_flags_list.add('KIVY_TARGET_PLATFORM_linux')
    elif 'android' == KIVY_TARGET_PLATFORM:
        macro_flags_list.add('KIVY_TARGET_PLATFORM_android')
    elif 'macosx' == KIVY_TARGET_PLATFORM:
        macro_flags_list.add('KIVY_TARGET_PLATFORM_macosx')
    elif 'ios' == KIVY_TARGET_PLATFORM:
        macro_flags_list.add('KIVY_TARGET_PLATFORM_ios')
    elif 'unknown' == KIVY_TARGET_PLATFORM:
        macro_flags_list.add('KIVY_TARGET_PLATFORM_unknown')
    
    if 'posix' == OS_API_TYPE:
        macro_flags_list.add('posix')
    elif 'nt' == OS_API_TYPE:
        macro_flags_list.add('nt')
    elif 'java' == OS_API_TYPE:
        macro_flags_list.add('java')
    
    if IS_RUNNING_IN_EMSCRIPTEN:
        macro_flags_list.add('IS_RUNNING_IN_EMSCRIPTEN')
    
    if IS_RUNNING_IN_WASI:
        macro_flags_list.add('IS_RUNNING_IN_WASI')
    
    if IS_RUNNING_IN_PYODIDE:
        macro_flags_list.add('IS_RUNNING_IN_PYODIDE')
    
    if IS_BUILDING_FOR_PYODIDE:
        macro_flags_list.add('IS_BUILDING_FOR_PYODIDE')
    
    if IS_INSIDE_OR_FOR_WEB_BROWSER:
        macro_flags_list.add('IS_INSIDE_OR_FOR_WEB_BROWSER')
    
    if IS_RUNNING_IN_PYCHARM:
        macro_flags_list.add('IS_RUNNING_IN_PYCHARM')

    macro_flags_list.update(additional_cflags)
    return macro_flags_list



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.data_manipulation.conversion.binary.versions.v_0.binary import *
from unittest import TestCase, main, skip


class TestBinaryConversions(TestCase):
    def test_bigint(self):
        data_values = [
            0,
            1,
            2,
            127,
            128,
            255,
            256,
            65535,
            65536,
            4294967295,
            4294967296,
            18446744073709551615,
            18446744073709551616,
            -0,
            -1,
            -2,
            -127,
            -128,
            -255,
            -256,
            -65535,
            -65536,
            -4294967295,
            -4294967296,
            -18446744073709551615,
            -18446744073709551616,
        ]

        for data_value in data_values:
            print(f'Testing: {data_value=}')
            self.assertEqual(data_value, bytes_to_bigint(bigint_to_bytes(data_value)))
    
    def test_ubigint(self):
        data_values = [
            0,
            1,
            2,
            127,
            128,
            255,
            256,
            65535,
            65536,
            4294967295,
            4294967296,
            18446744073709551615,
            18446744073709551616,
        ]

        for data_value in data_values:
            print(f'Testing: {data_value=}')
            self.assertEqual(data_value, bytes_to_ubigint(ubigint_to_bytes(data_value)))


if '__main__' == __name__:
    main()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['parse_accept_language', 'normalize_lang_from_parts', 'normalize_lang', 'optimize_accept_language', 
           'match_langs']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.text_processing.text_processing import removeprefix
from cengal.math.numbers import RationalNumber
from collections import OrderedDict
from typing import Mapping, Optional, List, Tuple, OrderedDict, Dict, Set


def parse_accept_language(headers: Mapping) -> Optional[OrderedDict[str, RationalNumber]]:
    accept_language_str = headers.get('Accept-Language'.casefold(), None)
    if accept_language_str is None:
        return None
    
    accept_lang_list = accept_language_str.split(',')
    result_list = list()
    for lang in accept_lang_list:
        lang: str = lang.strip()
        weight: Optional[RationalNumber] = None
        if ';' in lang:
            lang, weight = lang.split(';')
            lang = lang.strip()
            weight = weight.strip()
            weight = removeprefix(weight, 'q=')
            weight = weight.strip()
            try:
                weight = float(weight)
            except ValueError:
                weight = 0
        else:
            weight = 1
        
        result_list.append((lang, weight))
    
    result_list.sort(key=lambda x: x[1], reverse=True)
    return OrderedDict(result_list)


def normalize_lang_from_parts(main_lang: str, sub_lang: Optional[str]) -> str:
    if sub_lang:
        return f'{main_lang}-{sub_lang.upper()}'
    else:
        return main_lang


def normalize_lang(lang: str) -> str:
    if '-' in lang:
        main_lang, sub_lang = lang.split('-')
        main_lang = main_lang.strip()
        sub_lang = sub_lang.strip()
        lang = f'{main_lang}-{sub_lang.upper()}'
    else:
        lang = lang.strip()
    
    return lang


def optimize_accept_language(parsed_accept_language: Optional[OrderedDict[str, RationalNumber]]) -> Optional[OrderedDict[str, RationalNumber]]:
    if parsed_accept_language is None:
        return None
    
    result = OrderedDict()
    for lang, weight in parsed_accept_language.items():
        lang = lang.casefold()
        if '-' in lang:
            main_lang, sub_lang = lang.split('-')
            main_lang = main_lang.strip()
            sub_lang = sub_lang.strip()
            lang = f'{main_lang}-{sub_lang.upper()}'
            result[lang] = weight
            if main_lang in parsed_accept_language:
                pass
            else:
                result[main_lang] = weight
        else:
            result[lang] = weight
    
    return result


def match_langs(default_lang: str, 
                featured_langs: Set[str], 
                supported_langs: Set[str], 
                languages_mapping: Dict[str, str], 
                parsed_accept_language: Optional[OrderedDict[str, RationalNumber]]
                ) -> Tuple[str, OrderedDict[str, RationalNumber], OrderedDict[str, RationalNumber]]:
    better_lang: str = None
    featured_result = OrderedDict()
    result = OrderedDict()
    if parsed_accept_language is not None:
        for lang, weight in parsed_accept_language.items():
            if lang in languages_mapping:
                lang = languages_mapping[lang]
            
            if lang in featured_langs:
                featured_result[lang] = weight
            elif lang in supported_langs:
                result[lang] = weight
            else:
                pass
    
    if featured_result:
        for lang, weight in featured_result.items():
            better_lang = lang
            break
    elif result:
        for lang, weight in result.items():
            better_lang = lang
            break
    else:
        better_lang = default_lang

    return better_lang, featured_result, result



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['pycache_filter_func', 'remove_pycache_from_dir_tree']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


import shutil
import os
from typing import List, Tuple, Set
from cengal.file_system.path_manager import RelativePath, get_relative_path_part
from cengal.file_system.directory_manager import current_src_dir
from cengal.file_system.directory_manager import file_list_traversal, FilteringEntity


pycache_dir_name = '__pycache__'


def pycache_filter_func(filtering_entity: FilteringEntity, data) -> bool:
    if FilteringEntity.dirpath == filtering_entity:
        return True
    elif FilteringEntity.dirname == filtering_entity:
        dirpath, dirname = data
        if pycache_dir_name == dirname:
            return True
    elif FilteringEntity.aggregated == filtering_entity:
        new_data: List[Tuple[str, List[str], List[str]]] = list()
        for dirpath, dirnames, filenames in data:
            if dirnames:
                new_data.append((dirpath, dirnames, list()))
        
        return new_data
    else:
        return False


def remove_pycache_from_dir_tree(root_dir_path: str):
    to_be_deleted: Set[str] = set()
    for dirpath, dirnames, filenames in file_list_traversal(root_dir_path, pycache_filter_func):
        dirpath_rel: RelativePath = RelativePath(dirpath)
        for dirname in dirnames:
            pycache_full_path = dirpath_rel(dirname)
            to_be_deleted.add(pycache_full_path)
    
    for pycache_full_path in to_be_deleted:
        # print(f'Removing: {pycache_full_path}')
        if os.path.exists(pycache_full_path):
            shutil.rmtree(pycache_full_path, ignore_errors=True)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .net_io_abstract import *
import sys
import traceback

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class NetIO(NetIOBase):
    def __init__(self, transport):
        super().__init__(transport)

        self._need_to_stop = False
        self._already_begun = False

        self._new_connection_id = 0

        self.need_to_print_exceptions_info = False

    def destroy(self):
        self.method.destroy()

    def start(self, destroy_on_finish=True):
        if self._already_begun:
            raise LoopIsAlreadyBegun()

        self._already_begun = True
        try:
            while not self._need_to_stop:
                self.method.loop_iteration()
        finally:
            self._already_begun = False
            if destroy_on_finish:
                self.destroy()

    def stop(self):
        self._need_to_stop = True

    def make_connection(self, connection_info: ConnectionInfo=None, name=None)->Connection:
        new_connection = None
        if ConnectionType.passive == connection_info.connection_type:
            new_connection = self._make_passive_connection(connection_info, name)
        elif ConnectionType.active_connected == connection_info.connection_type:
            new_connection = self._make_active_connected_connection(connection_info, name)
        else:
            raise WrongConnectionType()
        return new_connection

    def add_connection(self, connection: Connection):
        self.all_connections.add(connection)
        if ConnectionType.passive == connection.connection_info.connection_type:
            self.passive_connections.add(connection)
        self.connection_by_id[connection.connection_id] = connection
        if connection.connection_name is not None:
            self.connection_by_name[connection.connection_name] = connection
        self.connection_by_fileno[connection.conn.fileno()] = connection

        if connection.worker_obj.api is None:
            connection.worker_obj.api = self
        if connection.worker_obj.connection is None:
            connection.worker_obj.connection = connection

        self.method.add_connection(connection.conn)
        self.check_is_connection_need_to_sent_data(connection)

    def remove_connection(self, connection: Connection):
        connection.connection_state = ConnectionState.waiting_for_disconnection
        self.method.set__should_be_closed(connection.conn)

    def on_accept_connection(self, connection):
        new_conn = None
        try:
            conn_and_address_pair = connection.conn.accept()
            new_conn, new_address = conn_and_address_pair
            new_connection = self._construct_active_accepted_connection(connection, conn_and_address_pair)
            self.add_connection(new_connection)
            try:
                new_connection.worker_obj.on_connect()
                self.check_is_connection_need_to_sent_data(new_connection)
            except:
                if __debug__: self.log_exception()
                self._set_connection_to_be_closed(new_connection, ConnectionState.worker_fault)
        except BlockingIOError:
            pass
        except:
            if __debug__: self.log_exception()
            if new_conn is not None:
                self.method.should_be_closed.add(new_conn)

    def on_connected(self, connection: Connection):
        connection.connection_state = ConnectionState.connected
        try:
            connection.worker_obj.on_connect()
            self.check_is_connection_need_to_sent_data(connection)
        except:
            if __debug__: self.log_exception()
            self._set_connection_to_be_closed(connection, ConnectionState.worker_fault)

    def on_read(self, connection: Connection):
        try:
            another_read_data_part = connection.conn.recv(1024)
            if another_read_data_part:
                connection.read_data += another_read_data_part
                try:
                    connection.worker_obj.on_read()
                    self.check_is_connection_need_to_sent_data(connection)
                except:
                    if __debug__: self.log_exception()
                    self._set_connection_to_be_closed(connection, ConnectionState.worker_fault)
            else:
                self._set_connection_to_be_closed(connection, ConnectionState.io_fault)
        except BlockingIOError:
            pass
        except:
            if __debug__: self.log_exception()
            self._set_connection_to_be_closed(connection, ConnectionState.io_fault)

    def on_write(self, connection: Connection):
        try:
            if connection.must_be_written_data:
                nsent = connection.conn.send(connection.must_be_written_data)
                connection.must_be_written_data = connection.must_be_written_data[nsent:]
            if not connection.must_be_written_data:
                try:
                    connection.worker_obj.on_no_more_data_to_write()
                    self.check_is_connection_need_to_sent_data(connection)
                except:
                    if __debug__: self.log_exception()
                    self._set_connection_to_be_closed(connection, ConnectionState.worker_fault)
        except BlockingIOError:
            pass
        except:
            if __debug__: self.log_exception()
            self._set_connection_to_be_closed(connection, ConnectionState.io_fault)

    def on_close(self, connection: Connection):
        self._remove_connection_from_internal_structures(connection)
        connection.conn.close()
        connection.connection_state = ConnectionState.disconnected
        try:
            connection.worker_obj.on_connection_lost()
        except:
            if __debug__: self.log_exception()
            pass

    def _get_new_connection_id(self):
        result = self._new_connection_id
        self._new_connection_id += 1
        return result

    def _construct_active_accepted_connection(self, base_passive_connection: Connection,
                                              conn_and_address_pair: tuple):
        conn, address = conn_and_address_pair
        new_worker_obj = copy.copy(base_passive_connection.worker_obj)
        new_worker_obj.api = None
        new_worker_obj.connection = None
        new_connection_info = ConnectionInfo(new_worker_obj,
                                             ConnectionType.active_accepted,
                                             address, conn.family, conn.type, conn.proto)
        new_connection = Connection(self._get_new_connection_id(), new_connection_info, conn_and_address_pair,
                                    ConnectionState.connected)
        return new_connection

    def _make_active_connected_connection(self, connection_info: ConnectionInfo=None, name=None)->Connection:
        conn = None
        try:
            conn = socket.socket(connection_info.socket_family, connection_info.socket_type,
                                 connection_info.socket_protocol, connection_info.socket_fileno)
            conn.setblocking(0)
            conn.connect(connection_info.socket_address)
        except (socket.error, OSError) as err:
            if err.errno not in {errno.EINPROGRESS, errno.EAGAIN}:
                conn.close()
                raise err
        conn_and_address_pair = (conn, connection_info.socket_address)
        new_connection = Connection(self._get_new_connection_id(), connection_info, conn_and_address_pair,
                                    ConnectionState.waiting_for_connection, name)
        self.add_connection(new_connection)
        self.method.set__need_write(new_connection.conn, True)
        return new_connection

    def _make_passive_connection(self, connection_info: ConnectionInfo=None, name=None)->Connection:
        conn = None
        try:
            conn = socket.socket(connection_info.socket_family, connection_info.socket_type,
                                 connection_info.socket_protocol, connection_info.socket_fileno)
            conn.setblocking(0)
            conn.bind(connection_info.socket_address)
            conn.listen(connection_info.backlog)
        except:
            conn.close()
            raise
        conn_and_address_pair = (conn, connection_info.socket_address)
        new_connection = Connection(self._get_new_connection_id(), connection_info, conn_and_address_pair,
                                    ConnectionState.connected, name)
        self.add_connection(new_connection)

        try:
            new_connection.worker_obj.on_connect()
            self.check_is_connection_need_to_sent_data(new_connection)
        except:
            if __debug__: self.log_exception()
            self._set_connection_to_be_closed(new_connection, ConnectionState.worker_fault)

        return new_connection

    def _remove_connection_from_internal_structures(self, connection: Connection):
        if connection in self.all_connections:
            self.all_connections.remove(connection)
        if connection in self.passive_connections:
            self.passive_connections.remove(connection)
        if connection.connection_id in self.connection_by_id:
            del self.connection_by_id[connection.connection_id]
        if connection.connection_name is not None:
            if connection.connection_name in self.connection_by_name:
                del self.connection_by_name[connection.connection_name]
        if connection.conn.fileno() in self.connection_by_fileno:
            del self.connection_by_fileno[connection.conn.fileno()]
        self.method.remove_connection(connection.conn)

    def _set_connection_to_be_closed(self, connection: Connection, state: ConnectionState):
        connection.connection_state = state
        self.method.set__should_be_closed(connection.conn)

    def check_is_connection_need_to_sent_data(self, connection: Connection):
        if connection.must_be_written_data or connection.force_write_call:
            if not isinstance(connection.must_be_written_data, memoryview):
                connection.must_be_written_data = memoryview(connection.must_be_written_data)
            self.method.set__need_write(connection.conn, True)
        else:
            self.method.set__need_write(connection.conn, False)

    def log_exception(self):
        if not self.need_to_print_exceptions_info:
            return
        exc = sys.exc_info()
        exception = exc
        error_str = '{} {}'.format(str(exception[0]), str(exception[1].args[0]))
        formatted_traceback = traceback.format_exception(exception[0], exception[1], exception[2])
        exception = exception[:2] + (formatted_traceback,)
        trace_str = ''.join(exception[2])
        result_string = '\n\tEXCEPTION:{}\n\tTRACE:{}'.format(error_str, trace_str)
        print(result_string)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
from os.path import expanduser
from cengal.text_processing.help_tools import get_text_in_brackets
import copy
from enum import Enum
from contextlib import contextmanager

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


# RESERVED_SYMBOLS__NAME = {'"', '='}
RESERVED_SYMBOLS__NAME = {'"'}
RESERVED_SYMBOLS__DATA = {'\n'}
RESERVED_SYMBOLS = RESERVED_SYMBOLS__NAME | RESERVED_SYMBOLS__DATA


class WrongSymbolInThePropertyName(Exception):
    pass


class WrongSymbolInThePropertyData(Exception):
    pass


class Errors(Enum):
    ok = 0
    file_not_found = 1


class ConfigManager:
    def __init__(self, app_tags, base_dir=None, default_content=None, immediate_save=True):
        """
        :param app_tags: ['Cute LLC', 'My App Name', 'configs for a parser']
        :param base_dir: User's Home if None
        :param immediate_save:
        :return:
        """
        self.app_tags = app_tags
        self.base_dir = base_dir or os.path.expanduser("~")
        self.config_path = os.path.join(self.base_dir, *self.app_tags)
        if not os.path.isdir(self.config_path):
            os.makedirs(self.config_path)
        self.default_content = default_content
        self.immediate_save = immediate_save

        self.files = dict()

        self.predefined_file_name = None

    @staticmethod
    def _filter_out_disallowed_symbols_from_property_name(property_name):
        for disallowed in RESERVED_SYMBOLS__NAME:
            property_name = property_name.replace(disallowed, '')
        return property_name

    def get_full_dir_path(self):
        return copy.copy(self.config_path)

    def load(self, file_name, force=None):
        if self.predefined_file_name is not None:
            force = file_name
            file_name = self.predefined_file_name
        # elif force is None:
        #     raise Exception('"force" property must be provided!')

        if force is None:
            force = False

        result = Errors.ok
        file_content = self.files.get(file_name)
        if (file_content is None) or force:
            # file_content = None
            file_path = os.path.join(self.config_path, file_name)
            file_lines = list()
            try:
                with open(file_path, 'r') as file:
                    file_lines = file.readlines()
            except FileNotFoundError as ex:
                result = Errors.file_not_found
            file_content = dict()
            for another_line in file_lines:
                another_line += '\n'
                key = get_text_in_brackets(another_line, '"', '"').strip()
                data = get_text_in_brackets(another_line, '=', '\n').strip()
                file_content[key] = data
            if file_name in self.default_content:
                file_default_content = self.default_content[file_name]
                for key, data in file_default_content.items():
                    if key not in file_content:
                        file_content[key] = data
            self.files[file_name] = file_content
            self.save(file_name)
        return result

    def save(self, file_name=None):
        if self.predefined_file_name is not None:
            file_name = self.predefined_file_name
        elif file_name is None:
            raise Exception('"file_name" property must be provided!')

        file_content = self.files.get(file_name)
        if file_content is not None:
            lines = list()
            for key, data in file_content.items():
                another_line = '"{}" = {}'.format(key, data)
                lines.append(another_line)
            new_file_content = '\n'.join(lines)
            file_path = os.path.join(self.config_path, file_name)
            with open(file_path, 'w') as file:
                file.write(new_file_content)

    def get_property(self, file_name, property_name=None):
        if self.predefined_file_name is not None:
            property_name = file_name
            file_name = self.predefined_file_name
        elif property_name is None:
            raise Exception('"property_name" property must be provided!')

        result = None
        self.load(file_name)
        file_content = self.files[file_name]
        property_name = self._filter_out_disallowed_symbols_from_property_name(property_name)
        result = copy.deepcopy(file_content.get(property_name))
        return result

    def set_property(self, file_name, property_name, data=None):
        """
        :param file_name:
        :param property_name:
        :param data: set to None to remove property from config
        :return:
        """

        if self.predefined_file_name is not None:
            data = property_name
            property_name = file_name
            file_name = self.predefined_file_name
        elif data is None:
            raise Exception('"data" property must be provided!')

        data = copy.deepcopy(data)
        self.load(file_name)
        file_content = self.files[file_name]
        if data is None:
            if property_name in file_content:
                del file_content[property_name]
        else:
            for disallowed in RESERVED_SYMBOLS__DATA:
                if disallowed in data:
                    raise WrongSymbolInThePropertyData(data)
            property_name = self._filter_out_disallowed_symbols_from_property_name(property_name)
            file_content[property_name] = copy.copy(data)
        if self.immediate_save:
            self.save(file_name)

    def remove_property(self, file_name, property_name=None):
        if self.predefined_file_name is not None:
            property_name = file_name
            file_name = self.predefined_file_name
        elif property_name is None:
            raise Exception('"property_name" property must be provided!')

        self.set_property(file_name, property_name, None)


@contextmanager
def conf_file_name(config_manager: ConfigManager, file_name: str):
    original_file_name = config_manager.predefined_file_name
    config_manager.predefined_file_name = file_name
    try:
        yield config_manager
    finally:
        config_manager.predefined_file_name = original_file_name
